from PIL import Image, ImageEnhance, ImageFilter, ImageDraw, ImageFont, ImageOps
import os
import requests
from io import BytesIO
from PIL import Image, ImageEnhance, ImageFilter, ImageDraw
import numpy as np
import random

class FotoFlow():
    def __init__(self, image_path_or_url):
        """بررسی اینکه آیا ورودی مسیر فایل محلی است یا لینک"""
        self.image_path_or_url = image_path_or_url
        
        if image_path_or_url.startswith("http"):
            response = requests.get(image_path_or_url)
            self.image = Image.open(BytesIO(response.content))
            print(f"Image loaded from URL: {image_path_or_url}")
        else:
            self.image = Image.open(image_path_or_url)
            print(f"Image loaded from local path: {image_path_or_url}")
        self._save_image()

    def show_image(self):
        """نمایش تصویر"""
        self.image.show()
        return self._get_image_bytes()

    def resize_image(self, width, height):
        """تغییر اندازه تصویر"""
        self.image = self.image.resize((width, height))
        print(f"Image resized to {width}x{height}")
        self._save_image()
        return self._get_image_bytes()

    def rotate_image(self, degrees):
        """چرخاندن تصویر"""
        self.image = self.image.rotate(degrees)
        print(f"Image rotated by {degrees} degrees")
        self._save_image()
        return self._get_image_bytes()

    def crop_image(self, left, top, right, bottom):
        """برش دادن تصویر"""
        self.image = self.image.crop((left, top, right, bottom))
        print(f"Image cropped with box: ({left}, {top}, {right}, {bottom})")
        self._save_image()
        return self._get_image_bytes()

    def apply_grayscale(self):
        """تبدیل تصویر به مقیاس خاکستری"""
        self.image = self.image.convert("L")
        print("Image converted to grayscale")
        self._save_image()
        return self._get_image_bytes()

    def apply_blur(self, radius=2):
        """اعمال فیلتر تاری به تصویر"""
        self.image = self.image.filter(ImageFilter.GaussianBlur(radius))
        print(f"Image blurred with radius {radius}")
        self._save_image()
        return self._get_image_bytes()

    def adjust_brightness(self, factor):
        """تنظیم روشنایی تصویر"""
        enhancer = ImageEnhance.Brightness(self.image)
        self.image = enhancer.enhance(factor)
        print(f"Image brightness adjusted by factor {factor}")
        self._save_image()
        return self._get_image_bytes()

    def adjust_contrast(self, factor):
        """تنظیم کنتراست تصویر"""
        enhancer = ImageEnhance.Contrast(self.image)
        self.image = enhancer.enhance(factor)
        print(f"Image contrast adjusted by factor {factor}")
        self._save_image()
        return self._get_image_bytes()

    def apply_edge_detection(self):
        """اعمال فیلتر شناسایی لبه‌ها"""
        self.image = self.image.filter(ImageFilter.FIND_EDGES)
        print("Edge detection applied")
        self._save_image()
        return self._get_image_bytes()

    def add_text(self, text, position, font_size=30):
        """افزودن متن به تصویر"""
        draw = ImageDraw.Draw(self.image)
        try:
            font = ImageFont.load_default()
        except IOError:
            font = ImageFont.truetype("arial.ttf", font_size)
        draw.text(position, text, font=font, fill="white")
        print(f"Text '{text}' added to image at position {position}")
        self._save_image()
        return self._get_image_bytes()

    def add_border(self, border_width, border_color=(0, 0, 0)):
        """افزودن حاشیه به تصویر"""
        width, height = self.image.size
        new_image = Image.new("RGB", (width + 2 * border_width, height + 2 * border_width), border_color)
        new_image.paste(self.image, (border_width, border_width))
        self.image = new_image
        print(f"Border of width {border_width} added to image")
        self._save_image()
        return self._get_image_bytes()

    def convert_to_transparent_png(self):
        """تبدیل تصویر به PNG با پس‌زمینه شفاف"""
        self.image = self.image.convert("RGBA")
        datas = self.image.getdata()

        new_data = []
        for item in datas:
            if item[0] in list(range(200, 256)) and item[1] in list(range(200, 256)) and item[2] in list(range(200, 256)):
                new_data.append((255, 255, 255, 0))
            else:
                new_data.append(item)
        self.image.putdata(new_data)
        print("Image converted to transparent PNG")
        self._save_image()
        return self._get_image_bytes()

    def apply_sepia(self):
        """اعمال فیلتر Sepia به تصویر"""
        width, height = self.image.size
        pixels = self.image.load()
        for py in range(height):
            for px in range(width):
                r, g, b = self.image.getpixel((px, py))

                tr = int(0.393 * r + 0.769 * g + 0.189 * b)
                tg = int(0.349 * r + 0.686 * g + 0.168 * b)
                tb = int(0.272 * r + 0.534 * g + 0.131 * b)

                if tr > 255:
                    tr = 255

                if tg > 255:
                    tg = 255

                if tb > 255:
                    tb = 255

                self.image.putpixel((px, py), (tr, tg, tb))
        print("Sepia filter applied")
        self._save_image()
        return self._get_image_bytes()

    def apply_invert(self):
        """اعمال فیلتر Invert (وارونه‌سازی رنگ‌ها)"""
        np_image = np.array(self.image)
        np_image = 255 - np_image
        self.image = Image.fromarray(np_image)
        print("Invert filter applied")
        self._save_image()
        return self._get_image_bytes()

    def apply_sharpen(self):
        """اعمال فیلتر تیز کردن تصویر"""
        self.image = self.image.filter(ImageFilter.SHARPEN)
        print("Sharpen filter applied")
        self._save_image()
        return self._get_image_bytes()

    def convert_to_webp(self):
        """تبدیل تصویر به فرمت WebP"""
        self.image.save(self._get_new_file_path("webp"))
        print("Image converted to WebP format")
        return self._get_image_bytes()

    def convert_to_bmp(self):
        """تبدیل تصویر به فرمت BMP"""
        self.image.save(self._get_new_file_path("bmp"))
        print("Image converted to BMP format")
        return self._get_image_bytes()

    def convert_to_gif(self):
        """تبدیل تصویر به فرمت GIF"""
        self.image.save(self._get_new_file_path("gif"))
        print("Image converted to GIF format")
        return self._get_image_bytes()

    def create_qr_code(self, data):
        """ایجاد کد QR از داده‌ها"""
        try:
            import qrcode
            qr = qrcode.make(data)
            self.image = qr
            print("QR Code generated")
            self._save_image()
            return self._get_image_bytes()
        except ModuleNotFoundError:
            import os
            try:
                os.system("pip install qrcode")
                return self.create_qr_code(data)
            except:
                raise ImportError("Not Model qrcode")

    def apply_mosaic(self, block_size):
        """افزودن افکت موزاییک به تصویر"""
        self.image = self.image.resize(
            (self.image.width // block_size, self.image.height // block_size),
            resample=Image.NEAREST
        ).resize(self.image.size, Image.NEAREST)
        print(f"Mosaic effect applied with block size {block_size}")
        self._save_image()
        return self._get_image_bytes()

    def _get_image_bytes(self):
        """تبدیل تصویر به بایت"""
        img_byte_arr = BytesIO()
        self.image.save(img_byte_arr, format='PNG')
        img_byte_arr.seek(0)
        return img_byte_arr.getvalue()

    def _save_image(self):
        """ذخیره تصویر به طور خودکار با همان نام قبلی یا نام فایل از URL"""
        if self.image_path_or_url.startswith("http"):
            file_name = os.path.basename(self.image_path_or_url)
            self.image.save(file_name)
            print(f"Image saved as {file_name}")
            return

        base_name = os.path.splitext(self.image_path_or_url)[0]
        file_extension = os.path.splitext(self.image_path_or_url)[1]
        self.image.save(f"{base_name}{file_extension}")
        print(f"Image saved as {base_name}{file_extension}")


    def _get_new_file_path(self, new_extension):
        """ایجاد مسیر جدید برای ذخیره تصویر با فرمت جدید"""
        base_name = os.path.splitext(self.image_path_or_url)[0]
        return f"{base_name}.{new_extension}"

    def apply_vignette(self):
        """اضافه کردن افکت وینیته به تصویر"""
        width, height = self.image.size
        pixels = self.image.load()

        center_x, center_y = width // 2, height // 2
        max_dist = np.sqrt(center_x ** 2 + center_y ** 2)

        for x in range(width):
            for y in range(height):
                dist = np.sqrt((x - center_x) ** 2 + (y - center_y) ** 2)
                factor = (max_dist - dist) / max_dist
                r, g, b = self.image.getpixel((x, y))
                r = int(r * factor)
                g = int(g * factor)
                b = int(b * factor)
                pixels[x, y] = (r, g, b)

        print("Vignette effect applied")
        return self._get_image_bytes()

    def apply_glow(self):
        """ایجاد افکت تابش نوری اطراف تصویر"""
        glow_image = self.image.filter(ImageFilter.GaussianBlur(5))
        glow_image = ImageEnhance.Brightness(glow_image).enhance(0.5)
        self.image.paste(glow_image, (0, 0), glow_image)
        print("Glow effect applied")
        return self._get_image_bytes()

    def apply_dodge(self):
        """ایجاد افکت داج (روشن کردن)"""
        img_array = np.array(self.image)
        img_array = np.minimum(255, img_array * 2)
        self.image = Image.fromarray(img_array)
        print("Dodge effect applied")
        return self._get_image_bytes()

    def apply_burn(self):
        """ایجاد افکت برن (تاریک کردن)"""
        img_array = np.array(self.image)
        img_array = np.maximum(0, img_array - 50)
        self.image = Image.fromarray(img_array)
        print("Burn effect applied")
        return self._get_image_bytes()

    def rotate_by_center(self, degrees):
        """چرخاندن تصویر حول مرکز آن"""
        self.image = self.image.rotate(degrees, resample=Image.BICUBIC, center=(self.image.width // 2, self.image.height // 2))
        print(f"Image rotated by {degrees} degrees around the center")
        return self._get_image_bytes()

    def mirror_image(self):
        """ایجاد تصویر آینه‌ای از تصویر اصلی"""
        self.image = self.image.transpose(Image.FLIP_LEFT_RIGHT)
        print("Image mirrored horizontally")
        return self._get_image_bytes()

    def flip_vertical(self):
        """معکوس کردن تصویر به صورت عمودی"""
        self.image = self.image.transpose(Image.FLIP_TOP_BOTTOM)
        print("Image flipped vertically")
        return self._get_image_bytes()

    def flip_horizontal(self):
        """معکوس کردن تصویر به صورت افقی"""
        self.image = self.image.transpose(Image.FLIP_LEFT_RIGHT)
        print("Image flipped horizontally")
        return self._get_image_bytes()

    def resize_by_percentage(self, percentage):
        """تغییر اندازه تصویر با درصد مشخص"""
        width, height = self.image.size
        new_width = int(width * percentage / 100)
        new_height = int(height * percentage / 100)
        self.image = self.image.resize((new_width, new_height))
        print(f"Image resized by {percentage}%")
        return self._get_image_bytes()

    def add_watermark(self, watermark_image):
        """اضافه کردن واترمارک به تصویر"""
        watermark = Image.open(watermark_image)
        self.image.paste(watermark, (self.image.width - watermark.width, self.image.height - watermark.height), watermark)
        print("Watermark added to image")
        return self._get_image_bytes()

    def convert_to_black_and_white(self):
        """تبدیل تصویر به سیاه و سفید"""
        self.image = self.image.convert("L")
        print("Image converted to black and white")
        return self._get_image_bytes()

    def apply_brightness_curve(self, curve):
        """اعمال منحنی روشنایی به تصویر"""
        img_array = np.array(self.image)
        img_array = np.clip(img_array * curve, 0, 255)
        self.image = Image.fromarray(img_array)
        print("Brightness curve applied")
        return self._get_image_bytes()

    def adjust_saturation(self, factor):
        """تنظیم اشباع رنگ‌ها"""
        enhancer = ImageEnhance.Color(self.image)
        self.image = enhancer.enhance(factor)
        print(f"Image saturation adjusted by factor {factor}")
        return self._get_image_bytes()

    def apply_negative(self):
        """ایجاد افکت منفی رنگ‌ها"""
        img_array = np.array(self.image)
        img_array = 255 - img_array
        self.image = Image.fromarray(img_array)
        print("Negative effect applied")
        return self._get_image_bytes()

    def generate_color_palette(self):
        """تولید پالت رنگ از تصویر"""
        img_array = np.array(self.image)
        unique_colors = np.unique(img_array.reshape(-1, img_array.shape[-1]), axis=0)
        print("Color palette generated")
        return unique_colors

    def create_collage(self, images):
        """ایجاد کولاژ از چندین تصویر"""
        widths, heights = zip(*(i.size for i in images))
        total_width = sum(widths)
        max_height = max(heights)
        collage = Image.new('RGB', (total_width, max_height))
        
        x_offset = 0
        for img in images:
            collage.paste(img, (x_offset, 0))
            x_offset += img.width
        
        self.image = collage
        print("Collage created")
        return self._get_image_bytes()

    def add_noise(self, amount):
        """افزودن نویز به تصویر"""
        img_array = np.array(self.image)
        noise = np.random.normal(0, amount, img_array.shape)
        img_array = np.clip(img_array + noise, 0, 255)
        self.image = Image.fromarray(img_array.astype(np.uint8))
        print(f"Noise added with amount {amount}")
        return self._get_image_bytes()

    def create_image_thumbnail(self):
        """ایجاد پیش‌نمایش از تصویر"""
        self.image.thumbnail((100, 100))
        print("Thumbnail created")
        return self._get_image_bytes()

    def resize_to_max_dimension(self, max_dimension):
        """تغییر اندازه تصویر بر اساس بزرگترین بعد"""
        width, height = self.image.size
        if width > height:
            new_width = max_dimension
            new_height = int(height * max_dimension / width)
        else:
            new_height = max_dimension
            new_width = int(width * max_dimension / height)
        self.image = self.image.resize((new_width, new_height))
        print(f"Image resized to max dimension {max_dimension}")
        return self._get_image_bytes()

    def draw_circle(self, position, radius):
        """رسم دایره روی تصویر"""
        draw = ImageDraw.Draw(self.image)
        draw.ellipse([position[0] - radius, position[1] - radius, position[0] + radius, position[1] + radius], outline="red")
        print(f"Circle drawn at {position} with radius {radius}")
        return self._get_image_bytes()

    def draw_rectangle(self, position, width, height):
        """رسم مستطیل روی تصویر"""
        draw = ImageDraw.Draw(self.image)
        draw.rectangle([position[0], position[1], position[0] + width, position[1] + height], outline="red")
        print(f"Rectangle drawn at {position} with width {width} and height {height}")
        return self._get_image_bytes()

    def apply_color_filter(self, color_filter):
        """اعمال فیلتر رنگی به تصویر"""
        img_array = np.array(self.image)
        img_array = img_array * color_filter
        self.image = Image.fromarray(np.clip(img_array, 0, 255).astype(np.uint8))
        print("Color filter applied")
        return self._get_image_bytes()

    def create_shadow(self, position, offset, blur_radius):
        """ایجاد سایه برای تصویر"""
        shadow = self.image.copy()
        shadow = shadow.convert("RGBA")
        shadow.putalpha(128)
        self.image.paste(shadow, (position[0] + offset, position[1] + offset), shadow)
        print(f"Shadow created with offset {offset} and blur radius {blur_radius}")
        return self._get_image_bytes()

    def apply_cartoon_effect(self):
        """ایجاد افکت کارتونی برای تصویر"""
        try :
            import cv2
            img_array = np.array(self.image)
            img_array = cv2.stylization(img_array, sigma_s=150, sigma_r=0.25)
            self.image = Image.fromarray(img_array)
            print("Cartoon effect applied")
            return self._get_image_bytes()
        except ModuleNotFoundError:
            import os
            try:
                os.system("pip install opencv-python")
                return self.apply_cartoon_effect()
            except:
                raise ImportError("Not Model cv2")

    def apply_sketch_effect(self):
        """تبدیل تصویر به طراحی دستی"""
        try:
            import cv2
            img_array = np.array(self.image)
            gray = cv2.cvtColor(img_array, cv2.COLOR_BGR2GRAY)
            invert = cv2.bitwise_not(gray)
            sketch = cv2.divide(gray, 255, invert)
            self.image = Image.fromarray(sketch)
            print("Sketch effect applied")
            return self._get_image_bytes()
        except ModuleNotFoundError:
            import os
            try:
                os.system("pip install opencv-python")
                return
            except:
                raise ImportError("Not Model cv2")

    def generate_painting_effect(self):
        """تبدیل تصویر به سبک نقاشی"""
        try:
            import cv2
            img_array = np.array(self.image)
            img_array = cv2.xphoto.createSimpleWB().balanceWhite(img_array)
            self.image = Image.fromarray(img_array)
            print("Painting effect applied")
            return self._get_image_bytes()
        except ModuleNotFoundError:
            import os
            try:
                os.system("pip install opencv-python")
                return
            except:
                raise ImportError("Not Model cv2")
    def apply_paint_splatter(self):
        """ایجاد افکت لکه‌های رنگی به تصویر"""
        img_array = np.array(self.image)
        for _ in range(50):
            x, y = random.randint(0, img_array.shape[1]), random.randint(0, img_array.shape[0])
            img_array[y, x] = [random.randint(0, 255) for _ in range(3)]
        self.image = Image.fromarray(img_array)
        print("Paint splatter effect applied")
        return self._get_image_bytes()

    def draw_line(self, start, end, width):
        """رسم خط روی تصویر"""
        draw = ImageDraw.Draw(self.image)
        draw.line([start, end], fill="red", width=width)
        print(f"Line drawn from {start} to {end}")
        return self._get_image_bytes()

    def blur_faces(self):
        """تار کردن چهره‌ها در تصویر"""
        try:
            import cv2
            face_cascade = cv2.CascadeClassifier(cv2.data.haarcascades + "haarcascade_frontalface_default.xml")
            img_array = np.array(self.image)
            gray = cv2.cvtColor(img_array, cv2.COLOR_BGR2GRAY)
            faces = face_cascade.detectMultiScale(gray, 1.3, 5)

            for (x, y, w, h) in faces:
                face = img_array[y:y+h, x:x+w]
                face = cv2.GaussianBlur(face, (99, 99), 30)
                img_array[y:y+h, x:x+w] = face

            self.image = Image.fromarray(img_array)
            print("Faces blurred in image")
            return self._get_image_bytes()
        except ModuleNotFoundError:
            import os
            try:
                os.system("pip install opencv-python")
                return
            except:
                raise ImportError("Not Model cv2")

    def extract_text_from_image(self):
        """استخراج متن از تصویر با استفاده از OCR"""
        try:
            import pytesseract
            text = pytesseract.image_to_string(self.image)
            print(f"Extracted text: {text}")
            return text
        except ModuleNotFoundError:
            import os
            try:
                os.system("pip install pytesseract")
                return
            except:
                raise ImportError("Not Model pytesseract")
    def _install_library(self, library_name):
        """چک کردن و نصب کتابخانه به طور خودکار در صورت نبودن"""
        try:
            __import__(library_name)
        except ModuleNotFoundError:
            print(f"{library_name} not found. Installing...")
            import os
            os.system(f"pip install {library_name}")
            try:
                __import__(library_name)
            except ImportError:
                raise ImportError(f"Failed to install {library_name}")

    def extract_text_from_image(self):
        """استخراج متن از تصویر با استفاده از OCR"""
        self._install_library("pytesseract")
        try:
            import pytesseract
            text = pytesseract.image_to_string(self.image)
            print(f"Extracted text: {text}")
            return text
        except Exception as e:
            print(f"Error in text extraction: {str(e)}")
            return None

    def apply_sharpen(self):
        """اعمال فیلتر تیزکننده به تصویر"""
        self._install_library("PIL")
        from PIL import ImageFilter
        try:
            self.image = self.image.filter(ImageFilter.SHARPEN)
            print("Image sharpened")
            return self._get_image_bytes()
        except Exception as e:
            print(f"Error in applying sharpen: {str(e)}")
            return None

    def apply_invert(self):
        """معکوس کردن رنگ‌های تصویر"""
        self._install_library("PIL")
        from PIL import ImageOps
        try:
            self.image = ImageOps.invert(self.image.convert("RGB"))
            print("Image colors inverted")
            return self._get_image_bytes()
        except Exception as e:
            print(f"Error in applying invert: {str(e)}")
            return None

    def rotate_image_center(self, degrees):
        """چرخاندن تصویر حول مرکز آن"""
        self._install_library("PIL")
        try:
            from PIL import Image
            self.image = self.image.rotate(degrees, resample=Image.BICUBIC, center=(self.image.width // 2, self.image.height // 2))
            print(f"Image rotated by {degrees} degrees")
            return self._get_image_bytes()
        except Exception as e:
            print(f"Error in rotating image: {str(e)}")
            return None

    def add_text_watermark(self, text, position=(0, 0), font_size=20):
        """افزودن واترمارک متنی به تصویر"""
        self._install_library("PIL")
        from PIL import ImageDraw, ImageFont
        try:
            draw = ImageDraw.Draw(self.image)
            font = ImageFont.load_default()
            draw.text(position, text, font=font, fill="white")
            print(f"Text watermark added: {text}")
            return self._get_image_bytes()
        except Exception as e:
            print(f"Error in adding text watermark: {str(e)}")
            return None

    def add_image_watermark(self, watermark_image_path, position=(0, 0)):
        """افزودن واترمارک تصویری به تصویر اصلی"""
        self._install_library("PIL")
        try:
            watermark = Image.open(watermark_image_path)
            self.image.paste(watermark, position, watermark)
            print("Image watermark added")
            return self._get_image_bytes()
        except Exception as e:
            print(f"Error in adding image watermark: {str(e)}")
            return None